from urllib.parse import urljoin

from parsel import Selector
from patul.request import Request


class Response(object):

    def __init__(self, response, request):
        self._resp = response
        self.text = self._resp.text
        self.request = request

    def __getattr__(self, item):
        return getattr(self._resp, item)

    def __str__(self):
        return f'<Response {self.status_code} {self.url}>'

    __repr__ = __str__

    def recode(self, encoding):
        self.text = self._resp.content.decode(encoding)
        self._init_selector()

    @property
    def meta(self):
        return self.request.meta

    @property
    def selector(self):
        if not hasattr(self, '_selector'):
            self._init_selector()
        return self._selector

    def _init_selector(self):
        self._selector = Selector(text=self.text)

    def xpath(self, query, **kw):
        return self.selector.xpath(query, **kw)

    def css(self, query):
        return self.selector.css(query)

    def urljoin(self, url):
        return urljoin(self.url, url)

    def follow(self, url, *args, **kwargs):
        if isinstance(url, Selector):
            url = url.xpath('./@href').get()
            if url is None:
                raise ValueError('selector does not have `href` attribute.')
        elif not isinstance(url, str):
            raise ValueError('param `url` only accept str or Selector type.')

        url = self.urljoin(url)
        return Request(url, *args, **kwargs)



from patul.crawler import Crawler, coro_wrapper
from patul.request import Request


class Spider(object):

    start_urls = None

    def __init__(self, loop):
        self.loop = loop

    async def __ainit__(self):
        pass

    def start_requests(self):
        for url in self.start_urls or []:
            yield Request(url)

    def parse(self, response):
        pass

    def process_result(self, result):
        pass

    def closed(self):
        pass


def crawl(spider_cls, close_loop=True, **kwargs):
    crawler = Crawler(**kwargs)
    loop = crawler.loop
    spider = spider_cls(loop)
    crawler.result_back = spider.process_result

    for request in spider.start_requests():
        if not request.callback:
            request.callback = spider.parse
        crawler.put_request(request)

    try:
        loop.run_until_complete(spider.__ainit__())
        crawler.run(close_loop=False)
    finally:
        try:
            coro = coro_wrapper(spider.closed())
            loop.run_until_complete(coro)
        finally:
            if close_loop:
                loop.close()



# coding=utf-8

'''
Request
'''
import random

from requests.cookies import RequestsCookieJar

USER_AGENTS = [
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:65.0) Gecko/20100101 Firefox/65.0',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/73.0.3683.103 Safari/537.36',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/64.0.3282.140 Safari/537.36 Edge/18.17763',
    'Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 10.0; WOW64; Trident/7.0; .NET4.0C; .NET4.0E; .NET CLR 2.0.50727; .NET CLR 3.0.30729; .NET CLR 3.5.30729)',
    'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/30.0.1599.101',
    'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/38.0.2125.122',
    'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.71',
    'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95',
    'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.1 (KHTML, like Gecko) Chrome/21.0.1180.71',
    'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; SV1; QQDownload 732; .NET4.0C; .NET4.0E)',
    'Mozilla/5.0 (Windows NT 5.1; U; en; rv:1.8.1) Gecko/20061208 Firefox/2.0.0 Opera 9.50',
    'Mozilla/5.0 (Windows NT 6.1; WOW64; rv:34.0) Gecko/20100101 Firefox/34.0',
]


class Request(object):

    def __init__(self, url, params=None, headers=None,
                 timeout=5, callback=None, meta=None,
                 cookies=None, proxies=None, method='GET', **kwargs):
        '''
        Request object
        :param url: request url
        '''
        self.url = url
        self.params = params
        self._headers = headers
        self._cookies = cookies
        self.timeout = timeout
        self.proxies = proxies
        self.method = method
        self.callback = callback
        self.meta = meta or {}
        kwargs.setdefault('allow_redirects', True)
        # it will as a kwargs send to requests.request
        self._kwargs = kwargs

    def requests_kwargs(self):
        """Use for requests"""
        return {
            'url': self.url,
            'params': self.params,
            'headers': self.headers,
            'cookies': self.cookies,
            'timeout': self.timeout,
            'proxies': self.proxies,
            'method': self.method,
            **self._kwargs
        }

    @property
    def headers(self):
        if self._headers is None:
            return {
                'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
                'Accept-Language': 'en',
                'User-Agent': random.choice(USER_AGENTS)
            }
        return self._headers

    @property
    def cookies(self):
        if isinstance(self._cookies, list):
            jar = RequestsCookieJar()
            for c in self._cookies:
                jar.set(c['name'], c['value'], domain=c['domain'], path=c['path'])
            return jar
        return self._cookies

    def __str__(self):
        return f'<Request {self.method} {self.url}>'


class FormRequest(Request):

    def __init__(self, url, data=None, json=None, method='POST', **kwargs):
        self.data = data
        self.json = json
        super().__init__(url=url, method=method, **kwargs)

    def requests_kwargs(self):
        kw = super().requests_kwargs()
        kw['data'] = self.data
        kw['json'] = self.json
        return kw



import asyncio
import concurrent.futures
import functools
import logging
from collections import Coroutine
from typing import Generator, AsyncGenerator

import requests as _requests
from requests.exceptions import ConnectionError, Timeout

from patul.request import Request
from patul.response import Response

LOG_FMT = '%(asctime)s [%(name)s] %(levelname)s %(message)s'
LOG_DATE_FMT = '%Y-%m-%d %H:%M:%S'


def _run_in_executor(func):
    @functools.wraps(func)
    def wrapped(crawler, *args):
        return crawler.loop.run_in_executor(crawler.executor, func, crawler, *args)
    return wrapped


class Crawler(object):

    def __init__(self,
                 start_requests=None,
                 result_back=None,
                 handle_cookies=True,
                 download_delay=0,
                 concurrent_requests=10,
                 max_retries=3,
                 priority=1,
                 log_settings=None,
                 loop=None):
        """
        :param result_back: function to process the result
        :param handle_cookies: handle the cookies or not
        :param download_delay: delayed time before download
        :param concurrent_requests: max concurrent requests
        :param priority: -1: first in first out, breadth-first
                          1: last in first out, depth-first
        :param log_settings: log settings
        :param loop: async event loop
        """
        if priority == -1:
            self._queue = asyncio.Queue()
        elif priority == 1:
            self._queue = asyncio.LifoQueue()
        else:
            raise ValueError(f'Argument priority expect 1 or -1, got {priority}')

        if handle_cookies:
            self.session = _requests.Session()
        else:
            self.session = _requests

        for request in start_requests or []:
            self.put_request(request)

        self.result_back = result_back
        self.download_delay = download_delay
        self.concurrent_requests = concurrent_requests
        self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=self.concurrent_requests)
        self.max_retries = max_retries
        self.loop = loop or asyncio.get_event_loop()
        self.logger = _logger(log_settings or {})

    def put_request(self, request):
        self._queue.put_nowait(request)

    def run(self, close_loop=True):
        self.loop.run_until_complete(self._run())
        self.close(close_loop)

    async def crawl(self, request):
        await asyncio.sleep(self.download_delay)
        response = await self.download(request)
        if response is not None:
            await self.process_response(response, request)

    async def _run(self):
        def iter_task():
            try:
                for _ in range(self.concurrent_requests):
                    request = self._queue.get_nowait()
                    yield self.crawl(request)
            except asyncio.queues.QueueEmpty:
                pass

        while self._queue.qsize():
            await asyncio.gather(*iter_task())

    def close(self, close_loop=True):
        if isinstance(self.session, _requests.Session):
            self.session.close()
        if close_loop:
            self.loop.close()
        self.executor.shutdown()

    @_run_in_executor
    def download(self, request):
        try:
            r = self.session.request(**request.requests_kwargs())
            return Response(r, request)
        except Exception as e:
            if isinstance(e, (ConnectionError, Timeout,)):
                self.loop.create_task(self.retry_download(request, e))
            else:
                self._log_err('downloading', request)

    async def retry_download(self, request, reason):
        retries = request.meta.get('retry_times', 0)
        if retries < self.max_retries:
            await self._queue.put(request)
            request.meta['retry_times'] = retries + 1
            self.logger.debug(
                f'Preparing to retry: {request} caused by {reason}. (retried {retries - 1} times)'
            )
        elif self.max_retries != 0:
            self.logger.debug(f'Gave up retry {request}')

    async def process_response(self, response, request):
        self.logger.debug(f'Downloaded: {response!r}')
        if not request.callback:
            return self.logger.warning(f'No function to parse {request}')
        try:
            outputs = request.callback(response)
            outputs = await coro_wrapper(outputs)
            await self.process_output(outputs)
        except:
            self._log_err('parsing', response)

    async def process_output(self, outputs):
        async for output in _iter_outputs(outputs):
            if output is None:
                continue
            if isinstance(output, Request):
                await self._queue.put(output)
            else:
                await self.process_result(output)

    async def process_result(self, result):
        self.logger.debug(f'Crawled result: {result}')
        if not self.result_back:
            return
        try:
            await coro_wrapper(self.result_back(result))
        except:
            self._log_err('processing result', result)

    def _log_err(self, situation, info):
        self.logger.error(f'Error happened when {situation}: {info}', exc_info=True)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


async def coro_wrapper(maybe_coro):
    if isinstance(maybe_coro, Coroutine):
        return await maybe_coro
    return maybe_coro


class AsyncIteratorWrapper:

    def __init__(self, obj):
        self._it = iter(obj)

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            return next(self._it)
        except StopIteration:
            raise StopAsyncIteration


def _iter_outputs(outputs):
    if isinstance(outputs, AsyncGenerator):
        return outputs
    if outputs is None:
        outputs = []
    elif not isinstance(outputs, Generator):
        outputs = [outputs]
    return AsyncIteratorWrapper(outputs)


def _logger(settings):
    logger = logging.getLogger('patul.crawler')
    logger.setLevel(settings.get('level', 'DEBUG'))

    if 'formatter' in settings:
        formatter = settings['formatter']
    else:
        formatter = logging.Formatter(fmt=LOG_FMT, datefmt=LOG_DATE_FMT)

    if 'fp' in settings:
        handler = logging.FileHandler(settings['fp'], encoding='utf-8')
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    if settings.get('to_stream', True):
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(formatter)
        logger.addHandler(stream_handler)

    return logger



from .crawler import Crawler
from .request import Request, FormRequest
from .spider import Spider, crawl


def fetch(url_or_request, **kwargs):
    crawler_kwargs = kwargs.pop('crawler_kwargs', {})
    crawler_kwargs.setdefault('max_retries', 0)
    crawler_kwargs.setdefault('concurrent_requests', 1)

    if isinstance(url_or_request, str):
        req = Request(url_or_request, **kwargs)
    elif isinstance(url_or_request, Request):
        req = url_or_request
    else:
        raise ValueError(f'unexpect type of param `url_or_request`: {type(url_or_request)}')

    resp = None

    if req.callback is None:
        def parse(response):
            nonlocal resp
            resp = response

        req.callback = parse

    with Crawler(**crawler_kwargs) as c:
        c.put_request(req)
        c.run(close_loop=False)

    return resp



import patul


class MySpider(patul.Spider):
    start_urls = ['https://www.baidu.com' for _ in range(1)]

    async def parse(self, response):
        print(response.url)
        yield patul.Request(response.url, callback=self.parse_next)

    def parse_next(self, response):
        print(response.url, ' from parse_next')
        yield {'hello': 'world'}

    def process_result(self, result):
        print(result)


if __name__ == '__main__':
    patul.crawl(MySpider)



# coding=utf-8

from setuptools import setup, find_packages

setup(
    name='patul',
    version=0.1005,
    description=(
        'A tiny spider based on asyncio and requests.'
    ),
    long_description_content_type='text/markdown',
    long_description=open('README.md').read(),
    author='financial',
    author_email='1012593988@qq.com',
    maintainer='financial',
    maintainer_email='1012593988@qq.com',
    license='BSD License',
    packages=find_packages(),
    platforms=["all"],
    url='https://github.com/financialfly/patul',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: Implementation',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries'
    ],
    install_requires=[
            'requests>=2.14.0',
            'parsel>=1.5.1'
        ]
)